###################################################################################################
# zmssqldb.py
#
# $Id: zmssqldb.py,v 1.3 2003/11/21 07:38:44 dnordmann Exp $
# $Name:  $
# $Author: dnordmann $
# $Revision: 1.3 $
#
# Implementation of class ZMSSqlDb (see below).
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
###################################################################################################

# Imports.
from __future__ import nested_scopes
from Globals import HTML, HTMLFile
from Products.ZSQLMethods.SQL import SQLConnectionIDs
import copy
import sys
import urllib
# Product Imports.
from zmsobject import ZMSObject
import _globals
import _objattrs


###################################################################################################
###################################################################################################
###   
###   C o n s t r u c t o r ( s )
###   
###################################################################################################
###################################################################################################

manage_addZMSSqlDbForm = HTMLFile('manage_addzmssqldbform', globals()) 
def manage_addZMSSqlDb(self, lang, manage_lang, _sort_id, REQUEST, RESPONSE):
  """ manage_addZMSSqlDb """
    
  ##### Create ####
  id_prefix = _globals.id_prefix(REQUEST.get('id','e'))
  obj = ZMSSqlDb(self.getNewId(id_prefix),_sort_id+1)
  self._setObject(obj.id, obj)
  
  obj = getattr(self,obj.id)
  ##### Object State ####
  obj.setObjStateNew(REQUEST)
  ##### Init Properties ####
  obj.manage_changeProperties(lang,manage_lang,REQUEST,RESPONSE)
  ##### VersionManager ####
  obj.onChangeObj(REQUEST)
            
  ##### Normalize Sort-IDs ####
  self.normalizeSortIds(id_prefix)
        
  # Return with message.        
  if REQUEST.RESPONSE:
    message = self.getLangStr('MSG_INSERTED',manage_lang)%obj.display_type(REQUEST)
    REQUEST.RESPONSE.redirect('%s/%s/manage_main?lang=%s&manage_lang=%s&manage_tabs_message=%s'%(self.absolute_url(),obj.id,lang,manage_lang,urllib.quote(message)))


###################################################################################################
###################################################################################################
###   
###   C l a s s
###   
###################################################################################################
###################################################################################################

class ZMSSqlDb(ZMSObject):

    # Properties.
    # -----------
    meta_type = "ZMSSqlDb"
    icon = "misc_/zms/zmssqldb.gif"
    icon_disabled = "misc_/zms/zmssqldb.gif"

    # Management Options.
    # -------------------
    manage_options = ( 
	{'label': 'TAB_EDIT',		'action': 'manage_main'},
	{'label': 'TAB_CONFIGURATION',	'action': 'manage_properties'},
	)

    # Management Permissions.
    # -----------------------
    __authorPermissions__ = (
		'manage','manage_main','manage_workspace',
		'manage_properties','manage_changeProperties',
		'manage_moveObjUp','manage_moveObjDown',
		'manage_cutObjects','manage_copyObjects','manage_pasteObjs',
		'manage_userForm','manage_user',
		)
    __editorPermissions__ = (
		)
    __administratorPermissions__ = (
		)
    __ac_permissions__=(
		('ZMS Author', __authorPermissions__),
		('ZMS Editor', __editorPermissions__),
		('ZMS Administrator', __administratorPermissions__),
		)

    # Properties.
    # -----------
    __obj_attrs__ = {
        # Changed by
        'change_uid':{'datatype':'string','multilang':1,'xml':0,'lang_inherit':0},
        'change_dt':{'datatype':'datetime','multilang':1,'xml':0,'lang_inherit':0},
        # Properties
        'sort_id':{'datatype':'string','xml':0},
    }


    # Management Interface.
    # ---------------------
    input_form = HTMLFile('dtml/zmssqldb/input_form', globals())
    intersection_sql = HTMLFile('dtml/zmssqldb/intersection_sql', globals())
    manage_main = HTMLFile('dtml/zmssqldb/manage_main', globals())
    manage_properties = HTMLFile('dtml/zmssqldb/manage_properties', globals())


    """
    ###############################################################################################    
    #
    #   CONSTRUCTOR
    #
    ###############################################################################################    
    """

    # -------------------------------------------------------------------------------------------------
    # ZMSSqlDb.recurse_updateVersion:
    #
    # Update version.
    # -------------------------------------------------------------------------------------------------
    def recurse_updateVersion(self, REQUEST):
      message = ''

      # Process super-objects.
      message += ZMSObject.recurse_updateVersion(self,REQUEST)
        
      # Process object.
      # [...]

      # Return with message.
      return message


    # ------------------------------------------------------------------------------------------
    #  ZMSSqlDb.sql_quote__:
    # ------------------------------------------------------------------------------------------
    def sql_quote__(self, tablename, columnname, v):
      entities = self.getEntities()
      entity = filter(lambda x: x['id'] == tablename, entities)[0]
      col = filter(lambda x: x['id'] == columnname , entity['columns'])[0]
      if col['type'] == 'int':
        try:
          return str(int(str(v)))
	except:
          return "NULL"
      elif col['type'] == 'float':
        try:
          return str(float(str(v)))
	except:
          return "NULL"
      else:
        v = str(v)
        if v.find("\'") >= 0: v=''.join(v.split("\'"),"''")
        return "'%s'"%v

    # ------------------------------------------------------------------------------------------
    #  ZMSSqlDb.query:
    # ------------------------------------------------------------------------------------------
    def query(self, qs):
      from cStringIO import StringIO
      from Shared.DC.ZRDB.Results import Results
      from Shared.DC.ZRDB import RDB
      databaseAdptr = getattr(self,self.connection_id)
      dbc = databaseAdptr._v_database_connection
      res = dbc.query(qs)
      if type(res) is type(''):
        f=StringIO()
        f.write(res)
        f.seek(0)
        result = RDB.File(f)
      else:
        result = Results(res)
      columns = []
      for result_column in result._searchable_result_columns():
        colName = result_column['name']
        colLabel = ''
        for s in colName.split('_'):
          colLabel += s.capitalize()
        try:
          colType = {'i':'int','t':'string','d':'datetime'}[result_column['type']]
        except:
          _globals.writeException(self,'[query]: Column ' + colName + ' has unknown type ' + str(result_column.get('type',None)) + '!')
        column = {}
        column['id'] = colName
        column['key'] = colName
        column['label'] = colLabel
        column['name'] = colLabel
        column['type'] = colType
        column['sort'] = 1
        columns.append(column)
      keys = map(lambda x: x['id'], columns)
      records = result
      # records = map(lambda x: _globals.map_key_vals(keys, x), result)
      return {'columns':columns,'records':records}


    # ------------------------------------------------------------------------------------------
    #  ZMSSqlDb.executeQuery:
    # ------------------------------------------------------------------------------------------
    def executeQuery(self, qs):
      from cStringIO import StringIO
      from Shared.DC.ZRDB.Results import Results
      from Shared.DC.ZRDB import RDB
      databaseAdptr = getattr(self,self.connection_id)
      dbc = databaseAdptr._v_database_connection
      res = dbc.query(qs)
      if type(res) is type(''):
        f=StringIO()
        f.write(res)
        f.seek(0)
        result=RDB.File(f)
      else:
        result=Results(res)
      return len(result)


    # ------------------------------------------------------------------------------------------
    #  ZMSSqlDb.getEntity:
    # ------------------------------------------------------------------------------------------
    def getEntity(self, tableName):
      return filter(lambda x: x['id'] == tableName, self.getEntities())[0]


    # ------------------------------------------------------------------------------------------
    #  ZMSSqlDb.getEntities:
    # ------------------------------------------------------------------------------------------
    def getEntities(self):
      entities = []
      model = getattr(self,'model',[])
      databaseAdptr = getattr(self,self.connection_id)
      for tableBrwsr in databaseAdptr.tpValues():
        tableName = tableBrwsr.Name().lower()
        tableType = tableBrwsr.Type().upper()
        if tableType == 'TABLE':
          # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
	  # +- COLUMNS
          # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
          cols = []
	  colIds = []
          for columnBrwsr in tableBrwsr.tpValues():
            colId = columnBrwsr.tpId().lower()
            colDescr = columnBrwsr.Description().upper()
	    colType = 'string'
	    colSize = None
            if colDescr.find('INT') >= 0:
              colType = 'int'
            elif colDescr.find('DATE') >= 0 or \
                 colDescr.find('TIME') >= 0:
              colType = 'datetime'
            elif colDescr.find('CHAR') >= 0:
              colSize = 50
              i = colDescr.find('(')
              if i >= 0:
                j = colDescr.find(')')
                if j >= 0:
                  colSize = int(colDescr[i+1:j])
              if colSize > 50:
                colType = 'text'
              else:
                colType = 'string'
            col = {}
            col['key'] = colId
            col['id'] = col['key']
            col['label'] = ''
            for s in colId.split('_'):
              col['label'] += s.capitalize()
            col['name'] = col['label']
            col['mandatory'] = colDescr.find('NOT NULL')>0
            col['type'] = colType
            col['sort'] = 1
	    # Custom Column Properties.
	    for modelTable in filter(lambda x: x['id']==tableName, model):
	      for modelTableCol in filter(lambda x: x['id']==colId, modelTable.get('columns',[])):
                for modelTableColProp in filter(lambda x: x not in ['id'], modelTableCol.keys()):
                  col[modelTableColProp] = modelTableCol[modelTableColProp]
            # Add Column.
            cols.append(col)
	    colIds.append(colId)
	  # Add Custom Columns.
          for modelTable in filter(lambda x: x['id']==tableName, model):
	    for modelTableCol in filter(lambda x: x['id'] not in colIds, modelTable.get('columns',[])):
              col = modelTableCol
              col['id'] = col.get('id','?')
	      col['index'] = col.get('index',len(cols))
	      col['type'] = col.get('type','?')
              col['key'] = col.get('key',col.get('id'))
              col['label'] = col.get('label',col.get('id'))
              cols.insert(col['index'], col)
	      colIds.append(col['id'])
          # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
	  # +- TABLE
          # +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-
          entity = {}
          entity['id'] = tableName
          entity['label'] = ''
          for s in tableName.split('_'):
            entity['label'] += s.capitalize()
          entity['columns'] = cols
          entity['type'] = 'table'
	  # Custom Table Properties.
          for modelTable in filter(lambda x: x['id']==tableName, model):
            for modelTableProp in filter(lambda x: x not in ['columns'], modelTable.keys()):
              entity[modelTableProp] = modelTable[modelTableProp]
          # Add Table.
          entities.append((entity['label'],entity))
      entities.sort()
      return map(lambda x: x[1], entities)
      

    """
    ###############################################################################################    
    ###
    ###   R e c o r d S e t
    ###
    ###############################################################################################    
    """

    # ---------------------------------------------------------------------------------------------
    #  ZMSSqlDb.recordSet_Init:
    #
    #  Initializes RecordSet.
    # ---------------------------------------------------------------------------------------------
    def recordSet_Init(self, REQUEST):
      SESSION = REQUEST.SESSION      
      tabledefs = self.getEntities()
      tablename = REQUEST.form.get('qentity',SESSION.get('qentity',tabledefs[0]['id']))
      tabledef = filter(lambda x: x['id']==tablename, tabledefs)[0]
      tablecols = tabledef['columns']
      #-- SELECT
      SESSION.set('qentity',tablename)
      sqlStatement = REQUEST.get('sqlStatement',[])
      sqlStatement.append('SELECT * FROM ' + tablename + ' ')
      REQUEST.set('sqlStatement',sqlStatement)
      # Columns
      REQUEST.set('grid_cols',tablecols)
      # Primary Key.
      primary_key = map(lambda x: x['id'], filter(lambda x: x.get('pk',0)==1, tablecols))
      primary_key.append(None)
      REQUEST.set('primary_key',primary_key[0])


    # ---------------------------------------------------------------------------------------------
    #  ZMSSqlDb.recordSet_Filter:
    #
    #  Filters RecordSet.
    # ---------------------------------------------------------------------------------------------
    def recordSet_Filter(self, REQUEST):
      SESSION = REQUEST.SESSION      
      tablename = SESSION['qentity']
      tabledefs = self.getEntities()
      tabledef = filter(lambda x: x['id']==tablename, tabledefs)[0]
      tablecols = tabledef['columns']
      #-- WHERE
      SESSION.set('qfilters',REQUEST.form.get('qfilters',SESSION.get('qfilters',1)))
      q = 'WHERE '
      for i in range(SESSION['qfilters']):
        filterattr='filterattr%i'%i
        filterop='filterop%i'%i
        filtervalue='filtervalue%i'%i
        sessionattr='%s_%s'%(filterattr,self.id)
        sessionop='%s_%s'%(filterop,self.id)
        sessionvalue='%s_%s'%(filtervalue,self.id)
        if REQUEST.get('action','')=='':
          if REQUEST.get('btn','')==self.getLangStr('BTN_RESET',REQUEST['manage_lang']):
            SESSION.set(sessionattr,'')
            SESSION.set(sessionop,'')
            SESSION.set(sessionvalue,'')
          elif REQUEST.get('btn','')==self.getLangStr('BTN_REFRESH',REQUEST['manage_lang']):
            SESSION.set(sessionattr,REQUEST.form.get(filterattr,''))
            SESSION.set(sessionop,REQUEST.form.get(filterop,''))
            SESSION.set(sessionvalue,REQUEST.form.get(filtervalue,''))
        for col in tablecols:
          columnname = col['id']
          v = SESSION.get(sessionvalue,'')
	  op = SESSION.get(sessionop,'=')
          if SESSION.get(sessionattr,'') == columnname and v != '':
            sqlStatement = REQUEST.get('sqlStatement',[])
            sqlStatement.append(q + columnname + ' ' + op + ' ' + self.sql_quote__(tablename, columnname, v) + ' ')
            REQUEST.set('sqlStatement',sqlStatement)
            q = 'AND '


    # ---------------------------------------------------------------------------------------------
    #  ZMSSqlDb.recordSet_Sort:
    #
    #  Sorts RecordSet.
    # ---------------------------------------------------------------------------------------------
    def recordSet_Sort(self, REQUEST):
      SESSION = REQUEST.SESSION      
      tablename = SESSION['qentity']
      tabledefs = self.getEntities()
      tabledef = filter(lambda x: x['id']==tablename, tabledefs)[0]
      tablecols = tabledef['columns']
      #-- ORDER BY
      qorder = REQUEST.get('qorder','')
      qorderdir = REQUEST.get('qorderdir','asc')
      if qorder == '' or not qorder in map(lambda x: x['id'], tablecols):
        for col in tablecols:
          if col.get('hide',0) != 1:
            qorder = col['id']
            if col.get('type','') in ['date','datetime']:
              qorderdir = 'desc'
	    break
      sqlStatement = REQUEST.get('sqlStatement',[])
      sqlStatement.append('ORDER BY ' + qorder + ' ' + qorderdir + ' ')
      REQUEST.set('sqlStatement',sqlStatement)
      REQUEST.set('qorder',qorder)
      REQUEST.set('qorderdir',qorderdir)


    """
    ###############################################################################################    
    ###
    ###   P r o p e r t i e s
    ###
    ###############################################################################################    
    """

    ###############################################################################################
    #  ZMSSqlDb.manage_changeProperties: 
    #
    #  Change Sql-Database properties.
    ###############################################################################################
    def manage_changeProperties(self, lang, manage_lang, REQUEST=None, RESPONSE=None): 
      """ ZMSSqlDb.manage_changeProperties """
      message = ''

      # Change.
      # -------
      if REQUEST.get('btn','') != self.getLangStr('BTN_CANCEL',manage_lang):
        self.connection_id = REQUEST['connection_id']
        self.max_rows = REQUEST['max_rows']
        self.model_xml = REQUEST['model']
        self.model = self.parseXmlString(REQUEST['model'])
        message = self.getLangStr('MSG_CHANGED',manage_lang)
      
      # Return with message.
      message = urllib.quote(message)
      return RESPONSE.redirect('manage_properties?lang=%s&manage_lang=%s&manage_tabs_message=%s'%(lang,manage_lang,message))


    # ---------------------------------------------------------------------------------------------
    #  ZMSSqlDb.isPage
    # ---------------------------------------------------------------------------------------------
    def isPage(self): return 0

    # ---------------------------------------------------------------------------------------------
    #  ZMSSqlDb.isPageElement
    # ---------------------------------------------------------------------------------------------
    def isPageElement(self): return 0

    # ---------------------------------------------------------------------------------------------
    #  ZMSSqlDb.getTitlealt
    # ---------------------------------------------------------------------------------------------
    def getTitlealt(self,REQUEST): return self.display_type(REQUEST)


    """
    ###############################################################################################
    ###
    ###  H T M L - P r e s e n t a t i o n 
    ###
    ###############################################################################################
    """

    # ---------------------------------------------------------------------------------------------
    #	ZMSSqlDb._getBodyContent:
    #
    #	HTML presentation of Textarea. 
    # ---------------------------------------------------------------------------------------------
    def _getBodyContent(self, REQUEST, thumbnail=0):
      # Render HTML presentation.
      text = '<h2>%s</h2>'%self.display_type(REQUEST)
      text += '<p class="description">'
      c = 0
      for entity in self.getEntities():
        if c > 0: text += '<br />'
        text += '- '+entity['id']
        c += 1
      text += '</p>'
      # Return <html>-presentation.
      return text

    # ---------------------------------------------------------------------------------------------
    #	ZMSSqlDb.renderShort:
    #
    #	Renders short presentation of Textarea.
    # ---------------------------------------------------------------------------------------------
    def renderShort(self, REQUEST):
      return self._getBodyContent(REQUEST,thumbnail=1)

###################################################################################################
